"""
Script to install Ghidra plugins.
"""

import argparse
from pathlib import Path

import pyhidra


if __name__ == "__main__":
    # spin everything up to ensure all new plugins are installed and exit
    parser = argparse.ArgumentParser("Install Ghidra Plugins")
    parser.add_argument(
        "--install-dir",
        type=Path,
        default=None,
        dest="install_dir",
        metavar="",
        help="Path to Ghidra installation. "\
             "(defaults to the GHIDRA_INSTALL_DIR environment variable)"
    )
    args = parser.parse_args()
    pyhidra.DeferredPyhidraLauncher(install_dir=args.install_dir).start()




import dataclasses
from datetime import datetime
from pathlib import Path
import re

from pyhidra import __version__


MINIMUM_GHIDRA_VERSION = "10.3"
_APPLICATION_PATTERN = re.compile(r"^application\.(\S+?)=(.*)$")


@dataclasses.dataclass(frozen=True)
class ApplicationInfo:
    """
    Ghidra Application Properties
    """
    name: str
    version: str
    release_name: str
    revision_ghidra: str = ""
    build_date: str = ""
    build_date_short: str = ""
    layout_version: str = ""
    gradle_min: str = ""
    java_min: str = ""
    java_max: str = ""
    java_compiler: str = ""
    gradle_max: str = ""

    @classmethod
    def from_file(cls, file: Path):
        """
        Parses Ghidra's application.properties file from the provided path
        """
        valid_fields = {f.name for f in dataclasses.fields(cls)}
        kwargs = dict()
        for line in file.read_text(encoding="utf8").splitlines():
            match = _APPLICATION_PATTERN.match(line)
            if not match:
                continue
            attr = match.group(1).replace('.', '_').replace('-', '_')
            value = match.group(2)
            if attr in valid_fields:
                kwargs[attr] = value
        return cls(**kwargs)


@dataclasses.dataclass
class ExtensionDetails:
    """
    Python side ExtensionDetails
    """
    name: str
    description: str
    author: str
    createdOn: str = dataclasses.field(default_factory=lambda: str(datetime.now()))
    version: str = None
    plugin_version: str = "0.0.1"

    @classmethod
    def from_file(cls, ext_path: Path):
        valid_fields = {f.name for f in dataclasses.fields(cls)}
        def cast(key, value):
            return cls.__annotations__[key](value)
        lines = ext_path.read_text().splitlines()
        kwargs = {
            key: cast(key, value)
            for key, value in map(lambda l: l.split("="), lines)
            if key in valid_fields
        }
        return cls(**kwargs)

    def __repr__(self):
        return "\n".join(f"{key}={value}" for key, value in dataclasses.asdict(self).items())



import shutil
import tempfile
from pathlib import Path
from os import pathsep


COMPILER_OPTIONS = ["-target", "17", "-source", "17"]


def _to_jar_(jar_path: Path, root: Path):
    from java.io import ByteArrayOutputStream
    from java.util.jar import JarEntry, JarOutputStream

    out = ByteArrayOutputStream()
    with JarOutputStream(out) as jar:
        for p in root.glob("**/*.class"):
            p = p.resolve()
            jar.putNextEntry(JarEntry(str(p.relative_to(root).as_posix())))
            jar.write(p.read_bytes())
            jar.closeEntry()
    jar_path.write_bytes(out.toByteArray())


def java_compile(src_path: Path, jar_path: Path):
    """
    Compiles the provided Java source

    :param src_path: The path to the java file or the root directory of the java source files
    :param jar_path: The path to write the output jar to
    """

    from java.lang import System
    from java.nio.file import Path as JPath
    from javax.tools import StandardLocation, ToolProvider

    with tempfile.TemporaryDirectory() as out:
        outdir = Path(out).resolve()
        compiler = ToolProvider.getSystemJavaCompiler()
        fman = compiler.getStandardFileManager(None, None, None)
        cp = [JPath @ (Path(p)) for p in System.getProperty("java.class.path").split(pathsep)]
        fman.setLocationFromPaths(StandardLocation.CLASS_PATH, cp)
        if src_path.is_dir():
            fman.setLocationFromPaths(StandardLocation.SOURCE_PATH, [JPath @ (src_path.resolve())])
        fman.setLocationFromPaths(StandardLocation.CLASS_OUTPUT, [JPath @ (outdir)])
        sources = None
        if src_path.is_file():
            sources = fman.getJavaFileObjectsFromPaths([JPath @ (src_path)])
        else:
            glob = src_path.glob("**/*.java")
            sources = fman.getJavaFileObjectsFromPaths([JPath @ (p) for p in glob])

        task = compiler.getTask(None, fman, None, COMPILER_OPTIONS, None, sources)

        if not task.call():
            # errors printed to stderr
            return

        if jar_path.suffix == '.jar':
            jar_path.parent.mkdir(exist_ok=True, parents=True)
            _to_jar_(jar_path, outdir)
        else:
            shutil.copytree(outdir, jar_path, dirs_exist_ok=True)




from pathlib import Path

from jpype import JConversion, JClass


@JConversion("java.lang.String", instanceof=Path)
def pathToString(cls: JClass, path: Path):
    return cls(path.resolve().__str__())


@JConversion("java.io.File", instanceof=Path)
def pathToFile(cls: JClass, path: Path):
    return cls(path)



import builtins
from keyword import iskeyword
from typing import Mapping, Sequence
from rlcompleter import Completer
from types import CodeType, FunctionType, MappingProxyType, MethodType, ModuleType

from docking.widgets.label import GLabel
from ghidra.app.plugin.core.console import CodeCompletion
from java.awt import Color
from java.util import Arrays, Collections
from jpype import JPackage
from jpype.types import JDouble, JFloat, JInt, JLong, JShort


NoneType = type(None)

CLASS_COLOR = Color(0, 0, 255)
CODE_COLOR = Color(0, 64, 0)
FUNCTION_COLOR = Color(0, 128, 0)
INSTANCE_COLOR = Color(128, 0, 128)
MAP_COLOR = Color(64, 96, 128)
METHOD_COLOR = Color(0, 128, 128)
NULL_COLOR = Color(255, 0, 0)
NUMBER_COLOR = Color(64, 64, 64)
PACKAGE_COLOR = Color(128, 0, 0)
SEQUENCE_COLOR = Color(128, 96, 64)

_TYPE_COLORS = {
    type: CLASS_COLOR,
    CodeType: CODE_COLOR,
    FunctionType: FUNCTION_COLOR,
    dict: MAP_COLOR,
    MappingProxyType: MAP_COLOR,
    MethodType: METHOD_COLOR,
    NoneType: NULL_COLOR,
    int: NUMBER_COLOR,
    float: NUMBER_COLOR,
    complex: NUMBER_COLOR,
    JShort: NUMBER_COLOR,
    JInt: NUMBER_COLOR,
    JLong: NUMBER_COLOR,
    JFloat: NUMBER_COLOR,
    JDouble: NUMBER_COLOR,
    ModuleType: PACKAGE_COLOR,
    JPackage: PACKAGE_COLOR
}


class PythonCodeCompleter(Completer):
    """
    Code Completer for Ghidra's Python interpreter window
    """

    _BUILTIN_ATTRIBUTE = object()
    __slots__ = ('cmd',)

    def __init__(self, py_console):
        super().__init__(py_console.locals.get_static_view())
        self.cmd: str

    def _get_label(self, i: int) -> GLabel:
        match = self.matches[i].rstrip("()")
        label = GLabel(match)
        attr = self.namespace.get(match, PythonCodeCompleter._BUILTIN_ATTRIBUTE)
        if attr is PythonCodeCompleter._BUILTIN_ATTRIBUTE:
            if iskeyword(match.rstrip()):
                return label
            attr = builtins.__dict__.get(match, PythonCodeCompleter._BUILTIN_ATTRIBUTE)
            if attr is not PythonCodeCompleter._BUILTIN_ATTRIBUTE and not match.startswith("__"):
                attr = builtins.__dict__[match]
            else:
                return label
        color = _TYPE_COLORS.get(type(attr), PythonCodeCompleter._BUILTIN_ATTRIBUTE)
        if color is PythonCodeCompleter._BUILTIN_ATTRIBUTE:
            t = type(attr)
            if isinstance(t, Sequence):
                color = SEQUENCE_COLOR
            elif isinstance(t, Mapping):
                color = MAP_COLOR
            else:
                color = INSTANCE_COLOR
        label.setForeground(color)
        return label

    def _supplier(self, i: int) -> CodeCompletion:
        insertion = self.matches[i][len(self.cmd):]
        return CodeCompletion(self.cmd, insertion, self._get_label(i))

    def get_completions(self, cmd: str):
        """
        Gets all the possible CodeCompletion(s) for the provided cmd

        :param cmd: The code to complete
        :return: A Java List of all possible CodeCompletion(s)
        """
        try:
            self.cmd = cmd
            if self.complete(cmd, 0) is None:
                return Collections.emptyList()
            res = CodeCompletion[len(self.matches)]
            Arrays.setAll(res, self._supplier)
            return Arrays.asList(res)
        except:  # pylint: disable=bare-except
            return Collections.emptyList()



import contextlib
import ctypes
import itertools
import logging
import rlcompleter
import sys
import threading
from code import InteractiveConsole

from ghidra.app.plugin.core.console import CodeCompletion
from ghidra.app.plugin.core.interpreter import InterpreterConsole, InterpreterPanelService
from ghidra.framework import Application
from java.io import BufferedReader, InputStreamReader, PushbackReader
from java.lang import ClassLoader, Runnable, String
from java.util import Collections
from java.util.function import Consumer
from jpype import JClass, JImplements, JOverride
from resources import ResourceManager

from pyhidra.java.plugin.completions import PythonCodeCompleter
from pyhidra.script import PyGhidraScript


logger = logging.getLogger(__name__)


def _get_private_class(path: str) -> JClass:
    gcl = ClassLoader.getSystemClassLoader()
    return JClass(path, loader=gcl)


def _get_plugin_class() -> JClass:
    return _get_private_class("dc3.pyhidra.plugin.PyhidraPlugin")


def _get_provider_class() -> JClass:
    return _get_private_class("dc3.pyhidra.plugin.PyScriptProvider")


def _get_interpreter_class() -> JClass:
    return _get_private_class("dc3.pyhidra.plugin.interpreter.PyhidraInterpreterConnection")


def _set_field(cls, fname, value, obj=None):
    cls = cls.class_
    field = cls.getDeclaredField(fname)
    field.setAccessible(True)
    field.set(obj, value)
    field.setAccessible(False)


def _run_script(script):
    PyGhidraScript(script).run()


class PyConsole(InteractiveConsole):
    """
    Pyhidra Interactive Console
    """

    def __init__(self, py_plugin) -> None:
        super().__init__(locals=PyGhidraScript(py_plugin.script))
        appVersion = Application.getApplicationVersion()
        appName = Application.getApplicationReleaseName()
        self.banner = f"Python Interpreter for Ghidra {appVersion} {appName}\n"
        self.banner += f"Python {sys.version} on {sys.platform}"
        self._plugin = py_plugin
        console = py_plugin.service.createInterpreterPanel(py_plugin, False)
        self._console = console
        self._reader = PushbackReader(InputStreamReader(console.getStdin()))
        self._line_reader = BufferedReader(self._reader)
        self._out = console.getOutWriter()
        self._err = console.getErrWriter()
        self._writer = self._out
        self._thread = None
        self._script = self.locals._script
        state = self._script.getState()
        self._script.set(state, console.getOutWriter())

    def raw_input(self, prompt=''):
        self._console.setPrompt(prompt)
        c = self._reader.read()
        if c == -1:
            raise EOFError
        if c == ord('\n'):
            return '\n'
        self._reader.unread(c)
        return self._line_reader.readLine()

    def write(self, data: str):
        self._writer.write(String @ data)
        self._writer.flush()

    def dispose(self):
        """
        Release the console resources
        """
        self._console.dispose()

    def close(self):
        if self._thread is not None and self._thread.is_alive():

            # raise a SystemExit in the interpreter thread once it resumes
            # this will force this thread, and only this thread, to begin
            # cleanup routines, __exit__ functions, finalizers, etc. and exit
            exc = ctypes.py_object(SystemExit)
            ctypes.pythonapi.PyThreadState_SetAsyncExc(self._thread.ident, exc)

            # closing stdin will wake up any thread attempting to read from it
            # this is required for the join to complete
            self._console.getStdin().close()

            # if we timeout then io out of our control is blocking it
            # at this point we tried and it will complete properly once it stops blocking
            self._thread.join(timeout=1.0)

            # ditch the locals so the contents may be released
            self.locals = dict()

    def restart(self):
        self.close()

        # clear any existing output in the window and re-open the console input
        self._console.clear()

        # this resets the locals, and gets a new code compiler
        super().__init__(locals=PyGhidraScript(self._script))

        target = self.interact
        targs = {"banner": self.banner}
        self._thread = threading.Thread(target=target, name="Interpreter", kwargs=targs)
        self._thread.start()

    @contextlib.contextmanager
    def redirect_writer(self):
        self._writer = self._err
        try:
            yield
        finally:
            self._writer = self._out

    def showsyntaxerror(self, filename=None):
        with self.redirect_writer():
            super().showsyntaxerror(filename=filename)

    def showtraceback(self) -> None:
        with self.redirect_writer():
            super().showtraceback()

    @contextlib.contextmanager
    def _run_context(self):
        self._script.start()
        success = False
        try:
            with contextlib.redirect_stdout(self._out), contextlib.redirect_stderr(self._err):
                yield
                success = True
        finally:
            self._script.end(success)

    def runcode(self, code):
        with self._run_context():
            super().runcode(code)
        self._out.flush()
        self._err.flush()


@JImplements("dc3.pyhidra.plugin.interpreter.PyhidraInterpreterConnection")
class PyPhidraPlugin:
    """
    The Python side PyhidraPlugin
    """

    def __init__(self, plugin):
        if hasattr(self, '_plugin'):
            # this gets entered twice for some reason
            return
        self._plugin = plugin
        self._actions = None
        self._logged_completions_change = False
        plugin_cls = _get_plugin_class()
        _set_field(plugin_cls, "finalizer", Runnable @ self.dispose, plugin)
        self.console = PyConsole(self)
        self.completer = PythonCodeCompleter(self.console)
        _get_interpreter_class().initialize(self)

    @staticmethod
    def register():
        plugin = _get_plugin_class()
        provider = _get_provider_class()
        _set_field(plugin, "initializer", Consumer @ PyPhidraPlugin)
        _set_field(provider, "scriptRunner", Consumer @ _run_script)

    def _set_plugin(self, plugin):
        self._plugin = plugin

    def dispose(self):
        """
        Release the plugin resources
        """
        if self._actions is not None:
            for action in self._actions:
                action.dispose()
        self.console.dispose()

    @property
    def program(self):
        return self._plugin.getCurrentProgram()

    @property
    def script(self):
        return self._plugin.script

    @property
    def service(self):
        return self._plugin.getTool().getService(InterpreterPanelService.class_)

    @JOverride
    def getCompletions(self, *args):
        try:
            if len(args) == 2:
                line, pos = args
                line = line[:pos]
            else:
                # older versions of Ghidra don't have the `end` argument.
                line, = args
            return self.completer.get_completions(line)
        except Exception as e:
            if not self._logged_completions_change:
                self._logged_completions_change = True
                logger.exception(e)
            return Collections.emptyList()

    @JOverride
    def getIcon(self):
        return ResourceManager.loadImage("images/python.png")

    @JOverride
    def getTitle(self):
        return "Pyhidra"

    @JOverride
    def getConsole(self) -> InterpreterConsole:
        return self.console._console

    @JOverride
    def getPlugin(self):
        return self._plugin

    @JOverride
    def close(self):
        self.console.close()

    @JOverride
    def restart(self):
        self.console.restart()

    @JOverride
    def setActions(self, actions):
        if self._actions is not None:
            for action in self._actions:
                action.dispose()
        self._actions = actions



from pathlib import Path
import pytest
from pyhidra.__main__ import _get_parser, PyhidraArgs


PROJECT_NAME = "stub_name"


class TestArgParser:

    def parse(self, *args) -> PyhidraArgs:
        parser = _get_parser()
        parser_args = PyhidraArgs(parser)
        args = [str(arg) for arg in args]
        parser.parse_args(args, namespace=parser_args)
        return parser_args

    @pytest.fixture(autouse=True)
    def _test_root(self, shared_datadir: Path):
        self.test_root = shared_datadir

    @property
    def example_script(self) -> Path:
        return self.test_root / "example_script.py"

    @property
    def example_exe(self) -> Path:
        return self.test_root / "strings.exe"

    @property
    def ghost_script(self) -> Path:
        return self.test_root / "ghost_script.py"

    @property
    def ghost_exe(self) -> Path:
        return self.test_root / "ghost.exe"

    def test_no_args(self):
        args = self.parse()
        assert args.valid

    def test_verbose_flag(self):
        args = self.parse("-v")
        assert args.verbose is True

    def test_project_name(self):
        args = self.parse("--project-name", PROJECT_NAME)
        assert args.project_name == PROJECT_NAME
        assert args.binary_path is None
        assert args.script_path is None
        assert args.project_path is None

    def test_project_path(self):
        args = self.parse("--project-path", self.test_root)
        assert args.valid
        assert args.project_path == self.test_root
        assert args.binary_path is None
        assert args.script_path is None
        assert args.project_name is None

    def test_script(self):
        args = self.parse(self.example_script)
        assert args.valid
        assert args.script_path == self.example_script

    def test_non_existing_script(self):
        args = self.parse(self.ghost_script)
        assert args.valid is False
        assert args.script_path == self.ghost_script
        assert args.binary_path is None

    def test_binary(self):
        args = self.parse(self.example_exe)
        assert args.valid
        assert args.binary_path == self.example_exe

    def test_non_existing_binary(self):
        args = self.parse(self.ghost_exe)
        assert args.valid is False
        assert args.binary_path == self.ghost_exe

    def test_non_existing_binary_plus_script(self):
        args = self.parse(self.ghost_exe, self.example_script)
        assert args.valid is False
        assert args.binary_path == self.ghost_exe
        assert args.script_path == self.example_script

    def test_script_with_non_existing_binary_arg(self):
        args = self.parse(self.example_script, self.ghost_exe)
        assert args.valid
        assert args.binary_path is None
        assert args.script_path == self.example_script
        assert args.script_args == [str(self.ghost_exe)]

    def test_script_with_optional_args(self):
        args = self.parse(self.example_script, "--project-path", "-v", self.test_root)
        assert args.valid
        assert args.verbose is False
        assert args.script_path == self.example_script
        assert args.script_args == ["--project-path", "-v", str(self.test_root)]

    def test_script_with_positional_args(self):
        args = self.parse(
            self.example_script,
            self.test_root,
            self.example_script,
            self.ghost_script
        )
        assert args.valid
        assert args.verbose is False
        assert args.binary_path is None
        assert args.script_path == self.example_script
        script_args = [
            str(arg) for arg in (self.test_root, self.example_script, self.ghost_script)
        ]
        assert args.script_args == script_args

    def test_script_with_intermingled_args(self):
        args = self.parse(
            self.example_script,
            self.example_exe,
            "-v",
            self.test_root,
            "--project-path",
            self.ghost_exe
        )
        assert args.valid
        assert args.verbose is False
        assert args.script_path == self.example_script
        script_args = [
            str(self.example_exe),
            "-v", str(self.test_root),
            "--project-path",
            str(self.ghost_exe)
        ]
        assert args.script_args == script_args

    def test_binary_script_with_intermingled_args(self):
        args = self.parse(
            "--project-name",
            PROJECT_NAME,
            self.example_exe,
            self.example_script,
            self.ghost_exe,
            "-v",
            self.test_root,
            "--project-name",
            self.ghost_exe
        )
        assert args.valid
        assert args.verbose is False
        assert args.project_name == PROJECT_NAME
        assert args.binary_path == self.example_exe
        assert args.script_path == self.example_script
        script_args = [
            str(self.ghost_exe),
            "-v",
            str(self.test_root),
            "--project-name",
            str(self.ghost_exe)
        ]
        assert args.script_args == script_args

    def test_skip_analysis(self):
        args = self.parse(
            "--skip-analysis"
        )
        assert args.skip_analysis

    def test_default_analysis(self):
        args = self.parse()
        assert not args.skip_analysis



import importlib
import importlib.util
import inspect
import logging
import sys
import traceback
from collections.abc import ItemsView, KeysView
from importlib.machinery import ModuleSpec, SourceFileLoader
from pathlib import Path
from jpype import JClass, JImplementationFor
from typing import List


_NO_ATTRIBUTE = object()

_headless_interpreter = None


class _StaticMap(dict):

    __slots__ = ('script',)

    def __init__(self, script: "PyGhidraScript"):
        super().__init__()
        self.script = script

    def __getitem__(self, key):
        res = self.get(key, _NO_ATTRIBUTE)
        if res is not _NO_ATTRIBUTE:
            return res
        raise KeyError(key)

    def get(self, key, default=None):
        res = self.script.get_static(key)
        return res if res is not _NO_ATTRIBUTE else default

    def __iter__(self):
        yield from self.script

    def keys(self):
        return KeysView(self)

    def items(self):
        return ItemsView(self)


class _JavaProperty(property):

    def __init__(self, field):
        super().__init__()
        self._field = field

    def __get__(self, obj, cls):
        return self._field.fget(obj)

    def __set__(self, obj, value):
        self._field.fset(obj, value)


#pylint: disable=too-few-public-methods
@JImplementationFor("dc3.pyhidra.plugin.PythonFieldExposer")
class _PythonFieldExposer:

    #pylint: disable=no-member
    def __jclass_init__(self):
        exposer = JClass("dc3.pyhidra.plugin.PythonFieldExposer")
        if self.class_ == exposer:
            return
        try:
            for k, v in exposer.getProperties(self.class_).items():
                self._customize(k, _JavaProperty(v))
        # allowing any exception to escape here may cause the jvm to terminate
        # pylint: disable=bare-except
        except:
            logger = logging.getLogger(__name__)
            logger.error("Failed to add property customizations for %s", self, exc_info=1)


class _GhidraScriptModule:

    def __init__(self, spec: ModuleSpec):
        super().__setattr__("__dict__", spec.loader_state["script"])

    def __setattr__(self, attr, value):
        if hasattr(self, attr):
            raise AttributeError(f"readonly attribute {attr}")
        super().__setattr__(attr, value)


class _GhidraScriptLoader(SourceFileLoader):

    def __init__(self, script: "PyGhidraScript", spec: ModuleSpec):
        super().__init__(spec.name, spec.origin)
        spec.loader_state = {"script": script}

    def create_module(self, spec: ModuleSpec):
        return _GhidraScriptModule(spec)


# pylint: disable=missing-function-docstring
class PyGhidraScript(dict):
    """
    Python GhidraScript Wrapper
    """

    def __init__(self, jobj=None):
        super().__init__()
        if jobj is None:
            jobj = JClass("dc3.pyhidra.plugin.PyScriptProvider").PyhidraHeadlessScript()
        self._script = jobj

        global _headless_interpreter

        from ghidra.util import SystemUtilities

        if SystemUtilities.isInHeadlessMode() and _headless_interpreter is None:
            _headless_interpreter = jobj

        # ensure the builtin set takes precedence over GhidraScript.set
        super().__setitem__("set", set)

        super().__setitem__("__this__", self._script)

        # this is injected since Ghidra commit e66e72577ded1aeae53bcc3f361dfce1ecf6e24a
        super().__setitem__("this", self._script)

    def __missing__(self, k):
        attr = getattr(self._script, k, _NO_ATTRIBUTE)
        if attr is not _NO_ATTRIBUTE:
            return attr
        raise KeyError(k)

    def __getattr__(self, item):
        return getattr(self._script, item)

    def __setitem__(self, k, v):
        attr = inspect.getattr_static(self._script, k, _NO_ATTRIBUTE)
        if attr is not _NO_ATTRIBUTE and isinstance(attr, property):
            setattr(self._script, k, v)
        else:
            super().__setitem__(k, v)

    def __iter__(self):
        yield from super().__iter__()
        yield from dir(self._script)

    def get_static(self, key):
        res = self.get(key, _NO_ATTRIBUTE)
        if res is not _NO_ATTRIBUTE:
            return res
        return inspect.getattr_static(self._script, key, _NO_ATTRIBUTE)

    def get_static_view(self):
        return _StaticMap(self)

    def set(self, state, monitor, writer):
        """
        see GhidraScript.set
        """
        self._script.set(state, monitor, writer)

    def run(self, script_path: str = None, script_args: List[str] = None):
        """
        Run this GhidraScript

        :param script_path: The path of the python script
        :param script_args: The arguments for the python script
        """
        sf = self._script.getSourceFile()
        if sf is None and script_path is None:
            return
        if script_path is None:
            script_path = sf.getAbsolutePath()
            script_args = self._script.getScriptArgs()

        if script_args is None:
            script_args = []
        else:
            self._script.setScriptArgs(script_args)

        orig_argv = sys.argv
        script_root = str(Path(script_path).parent)

        # honor the python safe_path flag introduced in 3.11
        safe_path = bool(getattr(sys.flags, "safe_path", 0))
        try:
            # Temporarily set command line arguments.
            sys.argv = [script_path] + list(script_args)

            if not safe_path:
                # add the directory containing the script to the start of the path
                # this provides the same import behavior as if the script was run normally
                sys.path.insert(0, script_root)

            spec = importlib.util.spec_from_file_location('__main__', script_path)
            spec.loader = _GhidraScriptLoader(self, spec)
            m = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(m)
            # pylint: disable=bare-except
            except:
                # filter the traceback so that it stops at the script
                exc_type, exc_value, exc_tb = sys.exc_info()
                i = 0
                tb = traceback.extract_tb(exc_tb)
                for fs in tb:
                    if fs.filename == script_path:
                        break
                    i += 1
                ss = traceback.StackSummary.from_list(tb[i:])
                e = traceback.TracebackException(exc_type, exc_value, exc_tb)
                e.stack = ss
                self._script.printerr(''.join(e.format()))
        finally:
            sys.argv = orig_argv

            if not safe_path:
                sys.path.remove(script_root)


def get_current_interpreter():
    """
    Gets the underlying GhidraScript for the focused Pyhidra InteractiveConsole.
    This will always return None unless it is being access from a function
    called from within the interactive console.

    :return: The GhidraScript for the active interactive console.
    """

    try:
        from ghidra.util import SystemUtilities
        from ghidra.framework.main import AppInfo

        global _headless_interpreter

        if SystemUtilities.isInHeadlessMode():
            if _headless_interpreter is None:
                # one hasn't been created yet so make one now
                PyScriptProvider = JClass("dc3.pyhidra.plugin.PyScriptProvider")
                _headless_interpreter = PyScriptProvider.PyhidraHeadlessScript()
            return _headless_interpreter

        project = AppInfo.getActiveProject()
        if project is None:
            return None

        ts = project.getToolServices()
        tool = None
        for t in ts.getRunningTools():
            if t.getActiveWindow().isFocused():
                tool = t
                break

        if tool is None:
            return None

        for plugin in tool.getManagedPlugins():
            if plugin.name == 'PyhidraPlugin':
                return plugin.script

    except ImportError:
        return None



import contextlib
import importlib.metadata
import inspect
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import threading
from pathlib import Path
from typing import List, NoReturn, Tuple, Union

import jpype
from jpype import imports, _jpype
from importlib.machinery import ModuleSpec

from . import __version__
from .javac import java_compile
from .script import PyGhidraScript
from .version import ApplicationInfo, ExtensionDetails, MINIMUM_GHIDRA_VERSION

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def _silence_java_output(stdout=True, stderr=True):
    from java.io import OutputStream, PrintStream
    from java.lang import System
    out = System.out
    err = System.err
    null = PrintStream(OutputStream.nullOutputStream())

    # The user's Java SecurityManager might not allow this
    with contextlib.suppress(jpype.JException):
        if stdout:
            System.setOut(null)
        if stderr:
            System.setErr(null)

    try:
        yield
    finally:
        with contextlib.suppress(jpype.JException):
            System.setOut(out)
            System.setErr(err)


def _load_entry_points(group: str, *args):
    """
    Loads any entry point callbacks registered by external python packages.
    """
    entry_points = importlib.metadata.entry_points()
    if hasattr(entry_points, 'select'):
        entries = entry_points.select(group=group)
    else:
        entries = entry_points.get(group, None)
        if entries is None:
            return

    for entry in entries:
        name = entry.name
        callback = entry.load()
        try:
            # Give launcher to callback so they can edit vmargs, install plugins, etc.
            logger.debug(f"Calling {group} entry point: {name}")
            callback(*args)
        except Exception as e:
            logger.error(f"Failed to run {group} entry point {name} with error: {e}")


class _PyhidraImportLoader:
    """ (internal) Finder hook for importlib to handle Python mod conflicts. """

    def find_spec(self, name, path, target=None):

        # If jvm is not started then there is nothing to find.
        if not _jpype.isStarted():
            return None

        if name.endswith('_') and _jpype.isPackage(name[:-1]):
            return ModuleSpec(name, self)

    def create_module(self, spec):
        return _jpype._JPackage(spec.name[:-1])

    def exec_module(self, fullname):
        pass


class PyhidraLauncher:
    """
    Base pyhidra launcher
    """

    def __init__(self, verbose=False, *, install_dir: Path = None):
        """
        Initializes a new `PyhidraLauncher`.

        :param verbose: True to enable verbose output when starting Ghidra.
        :param install_dir: Ghidra installation directory.
            (Defaults to the GHIDRA_INSTALL_DIR environment variable)
        :raises ValueError: If the Ghidra installation directory is invalid.
        """
        self._layout = None
        self._java_home = None

        install_dir = install_dir or os.getenv("GHIDRA_INSTALL_DIR")
        self._install_dir = self._validate_install_dir(install_dir)
        self._plugins: List[Tuple[Path, ExtensionDetails]] = []
        self.verbose = verbose

        ghidra_dir = self._install_dir / "Ghidra"
        self.class_path = [str(ghidra_dir / "Framework" / "Utility" / "lib" / "Utility.jar")]
        self.class_files = []
        self.vm_args = self._jvm_args(self._install_dir)
        self.args = []
        self.app_info = ApplicationInfo.from_file(ghidra_dir / "application.properties")

    @classmethod
    def _jvm_args(cls, install_dir: Path) -> List[str]:
        suffix = "_" + platform.system().upper()
        option_pattern: re.Pattern = re.compile(fr"VMARGS(?:{suffix})?=(.+)")
        properties = []

        launch_properties = install_dir / "support" / "launch.properties"

        with open(launch_properties, "r", encoding='utf-8') as fd:
            # this file is small so just read it at once
            for line in fd.readlines():
                match = option_pattern.match(line)
                if match:
                    arg = match.group(1)
                    name, sep, value = arg.partition('=')
                    # unquote any values because quotes are automatically added during JVM startup
                    if value.startswith('"') and value.endswith('"'):
                        value = value.removeprefix('"').removesuffix('"')
                    properties.append(name + sep + value)
        return properties

    @property
    def extension_path(self) -> Path:
        if not self._layout:
            raise RuntimeError("extension_path cannot be obtained until launcher starts.")
        return Path(self._layout.getUserSettingsDir().getPath()) / "Extensions"

    @property
    def java_home(self) -> Path:
        if not self._java_home:
            launch_support = self.install_dir / "support" / "LaunchSupport.jar"
            if not launch_support.exists():
                raise ValueError(f"{launch_support} does not exist")
            cmd = f'java -cp "{launch_support}" LaunchSupport "{self.install_dir}" -jdk_home -save'
            home = subprocess.check_output(cmd, encoding="utf-8", shell=True)
            self._java_home = Path(home.rstrip())
        return self._java_home

    @java_home.setter
    def java_home(self, path: Path):
        self._java_home = Path(path)

    @property
    def install_dir(self) -> Path:
        return self._install_dir

    @classmethod
    def _validate_install_dir(cls, install_dir: Union[Path, str]) -> Path:
        """
        Validates and sets the Ghidra installation directory.
        """
        if not install_dir:
            msg = (
                "Please set the GHIDRA_INSTALL_DIR environment variable "
                "or `install_dir` during the Launcher construction to the "
                "directory where Ghidra is installed."
            )
            cls._report_fatal_error("GHIDRA_INSTALL_DIR is not set", msg, ValueError(msg))

        # both the directory and the application.properties file must exist
        install_dir = Path(install_dir)
        if not install_dir.exists():
            msg = f"{install_dir} does not exist"
            cls._report_fatal_error("Invalid Ghidra Installation Directory", msg, ValueError(msg))

        path = install_dir / "Ghidra" / "application.properties"
        if not path.exists():
            msg = "The Ghidra installation does not contain the required " + \
                  "application.properties file"
            cls._report_fatal_error("Corrupt Ghidra Installation", msg, ValueError(msg))

        return install_dir

    def add_classpaths(self, *args):
        """
        Add additional entries to the classpath when starting the JVM
        """
        self.class_path += args

    def add_vmargs(self, *args):
        """
        Add additional vmargs for launching the JVM
        """
        self.vm_args += args

    def add_class_files(self, *args):
        """
        Add additional entries to be added the classpath after Ghidra has been fully loaded.
        This ensures that all of Ghidra is available so classes depending on it can be properly loaded.
        """
        self.class_files += args

    @classmethod
    def _report_fatal_error(cls, title: str, msg: str, cause: Exception) -> NoReturn:
        logger.error("%s: %s", title, msg)
        raise cause

    def check_ghidra_version(self):
        """
        Checks if the currently installed Ghidra version is supported.
        The launcher will report the problem and terminate if it is not supported.
        """
        if self.app_info.version < MINIMUM_GHIDRA_VERSION:
            msg = f"Ghidra version {self.app_info.version} is not supported" + os.linesep + \
                  f"The minimum required version is {MINIMUM_GHIDRA_VERSION}"
            self._report_fatal_error("Unsupported Version", msg, ValueError(msg))

    def start(self, **jpype_kwargs):
        """
        Starts Jpype connection to Ghidra (if not already started).
        """
        if jpype.isJVMStarted():
            return

        self.check_ghidra_version()

        try:
            # Before starting up, give launcher to installed entry points so they can do their thing.
            _load_entry_points("pyhidra.setup", self)

            pyhidra_details = ExtensionDetails(
                name="pyhidra",
                description="Native Python Plugin",
                author="Department of Defense Cyber Crime Center (DC3)",
                plugin_version=__version__,
                version=self.app_info.version
            )

            # Merge classpath
            jpype_kwargs['classpath'] = self.class_path + jpype_kwargs.get('classpath', [])

            # force convert strings (required by pyhidra)
            jpype_kwargs['convertStrings'] = True

            # set the JAVA_HOME environment variable to the correct one so jpype uses it
            os.environ['JAVA_HOME'] = str(self.java_home)

            jpype.startJVM(
                None, # indicates to use JAVA_HOME as the jvm path
                *self.vm_args,
                **jpype_kwargs
            )

            # Install hook into python importlib
            sys.meta_path.append(_PyhidraImportLoader())

            imports.registerDomain("ghidra")

            # import and create a temporary GhidraApplicationLayout this can be
            # used without initializing Ghidra to obtain the correct Extension path
            from ghidra import GhidraApplicationLayout
            self._layout = GhidraApplicationLayout()

            # uninstall any outdated plugins before initializing Ghidra to ensure they are loaded correctly
            self._uninstall_old_plugin(pyhidra_details)

            for _, details in self._plugins:
                try:
                    self._uninstall_old_plugin(details)
                except:
                    logger.warning("failed to uninstall plugin %s", details.name)


            from ghidra import GhidraLauncher
            self._layout = GhidraLauncher.initializeGhidraEnvironment()

            # install the Pyhidra plugin.
            from pyhidra.java import plugin
            needs_reload = self._install_plugin(Path(plugin.__file__).parent, pyhidra_details)

            if needs_reload:
                # "restart" Ghidra
                self._layout = GhidraLauncher.initializeGhidraEnvironment()
                needs_reload = False

            from java.lang import System

            # manually check the classpath for the pyhidra plugin to
            # help diagnose confusing errors (GH #31)
            # this will help in the future too if Extensions are ever moved outside
            # of the Ghidra user settings directory
            jar_path = self._get_plugin_jar_path("pyhidra")

        except Exception as e:
            self._report_fatal_error("An error occured launching Ghidra", str(e), e)

        # NOTE: be very careful not to cause an exception here because there will be
        # no indication of a problem in GUI mode unless started with pyhidra -g -v

        CLASSPATH_PROPERTY = "java.class.path"
        classpath = System.getProperty(CLASSPATH_PROPERTY)
        if classpath is None:
            # this is impossible but a helpful message is
            # better than "'NoneType' object hsas no attribute 'split'"
            msg = f"Required Java property {CLASSPATH_PROPERTY} not found"
            self._report_fatal_error(f"No Classpath", msg, RuntimeError(msg))

        # ensure it is a Python string (not Java string) and then split the classpath
        classpath = str(classpath).split(os.pathsep)
        if str(jar_path) not in classpath:
            title = "Classpath Setup Error"
            msg = "Pyhidra plugin is not in the system classpath"
            # Ghidra uses this property depending on JVM configuration
            ext_classpath = System.getProperty("java.class.path.ext")
            if ext_classpath is None:
                logger.debug("plugin path: %s\nclasspath: %s", jar_path, '\n'.join(classpath))
                self._report_fatal_error(title, msg, RuntimeError(msg))
            classpath = str(ext_classpath).split(os.pathsep)
            if str(jar_path) not in classpath:
                logger.debug("plugin path: %s\nclasspath: %s", jar_path, '\n'.join(classpath))
                self._report_fatal_error(title, msg, RuntimeError(msg))

        try:
            # import it at the end so interfaces in our java code may be implemented
            from pyhidra.java.plugin.plugin import PyPhidraPlugin
            PyPhidraPlugin.register()

            # Add extra class paths
            # Do this before installing plugins incase dependencies are needed
            if self.class_files:
                from java.lang import ClassLoader
                gcl = ClassLoader.getSystemClassLoader()
                for path in self.class_files:
                    gcl.addPath(path)

            # Install extra plugins.
            for source_path, details in self._plugins:
                needs_reload = self._install_plugin(source_path, details) or needs_reload

            if needs_reload:
                # "restart" Ghidra
                self._layout = GhidraLauncher.initializeGhidraEnvironment()

            # import properties to register the property customizer
            from . import properties as _

            _load_entry_points("pyhidra.pre_launch")

            self._launch()
        except Exception as e:
            self._report_fatal_error("An error occured launching Ghidra", str(e), e)

    def get_install_path(self, plugin_name: str) -> Path:
        """
        Obtains the path for installation of a given plugin.
        """
        return self.extension_path / plugin_name

    def _get_plugin_jar_path(self, plugin_name: str) -> Path:
        return self.get_install_path(plugin_name) / "lib" / (plugin_name + ".jar")

    def uninstall_plugin(self, plugin_name: str):
        """
        Uninstalls given plugin.
        """
        path = self.get_install_path(plugin_name)
        if path.exists():
            # delete the existing extension so it will be up-to-date
            try:
                shutil.rmtree(path)
            except Exception as e:
                self._report_fatal_error(
                    "Plugin Update Failed",
                    f"Could not delete existing plugin at\n{path}",
                    e
                )

    def _uninstall_old_plugin(self, details: ExtensionDetails):
        """
        Automatically uninstalls an outdated plugin if it exists.
        """
        plugin_name = details.name
        path = self.get_install_path(plugin_name)
        ext = path / "extension.properties"

        # Uninstall old version.
        if path.exists() and ext.exists():
            orig_details = ExtensionDetails.from_file(ext)
            if not orig_details.plugin_version or orig_details.plugin_version != details.plugin_version:
                self.uninstall_plugin(plugin_name)
                logger.info(f"Uninstalled older plugin: {plugin_name} {orig_details.plugin_version}")

    def _install_plugin(self, source_path: Path, details: ExtensionDetails):
        """
        Compiles and installs a Ghidra extension if not already installed.
        """
        if details.version is None:
            details.version = self.app_info.version
        plugin_name = details.name
        path = self.get_install_path(plugin_name)
        ext = path / "extension.properties"
        manifest = path / "Module.manifest"
        root = source_path

        if not path.exists():
            path.mkdir(parents=True)

            jar_path = path / "lib" / (plugin_name + ".jar")
            java_compile(root.parent, jar_path)

            ext.write_text(str(details))

            # required empty file
            manifest.touch()

            # Copy over ghidra_scripts if included.
            ghidra_scripts = root / "ghidra_scripts"
            if ghidra_scripts.exists():
                shutil.copytree(ghidra_scripts, path / "ghidra_scripts")

            logger.info(f"Installed plugin: {plugin_name} {details.plugin_version}")
            return True

        return False

    def install_plugin(self, source_path: Path, details: ExtensionDetails):
        """
        Compiles and installs a Ghidra extension when launcher is started.
        """
        self._plugins.append((source_path, details))

    def _launch(self):
        pass

    @staticmethod
    def has_launched() -> bool:
        """
        Checks if jpype has started and if Ghidra has been fully initialized.
        """
        if not jpype.isJVMStarted():
            return False

        from ghidra.framework import Application
        return Application.isInitialized()


class DeferredPyhidraLauncher(PyhidraLauncher):
    """
    PyhidraLauncher which allows full Ghidra initialization to be deferred.
    initialize_ghidra must be called before all Ghidra classes are fully available.
    """

    def initialize_ghidra(self, headless=True):
        """
        Finished Ghidra initialization

        :param headless: whether or not to initialize Ghidra in headless mode.
            (Defaults to True)
        """
        from ghidra import GhidraRun
        from ghidra.framework import Application, HeadlessGhidraApplicationConfiguration
        with _silence_java_output(not self.verbose, not self.verbose):
            if headless:
                config = HeadlessGhidraApplicationConfiguration()
                Application.initializeApplication(self._layout, config)
            else:
                GhidraRun().launch(self._layout, self.args)


class HeadlessPyhidraLauncher(PyhidraLauncher):
    """
    Headless pyhidra launcher
    """

    def _launch(self):
        from ghidra.framework import Application, HeadlessGhidraApplicationConfiguration
        with _silence_java_output(not self.verbose, not self.verbose):
            config = HeadlessGhidraApplicationConfiguration()
            Application.initializeApplication(self._layout, config)


class _PyhidraStdOut:

    def __init__(self, stream):
        self._stream = stream

    def _get_current_script(self) -> "PyGhidraScript":
        for entry in inspect.stack():
            f_globals = entry.frame.f_globals
            if isinstance(f_globals, PyGhidraScript):
                return f_globals

    def flush(self):
        script = self._get_current_script()
        if script is not None:
            writer = script._script.writer
            if writer is not None:
                writer.flush()
                return

        self._stream.flush()

    def write(self, s: str) -> int:
        script = self._get_current_script()
        if script is not None:
            writer = script._script.writer
            if writer is not None:
                writer.write(s)
                return len(s)

        return self._stream.write(s)


class GuiPyhidraLauncher(PyhidraLauncher):
    """
    GUI pyhidra launcher
    """

    @classmethod
    def popup_error(cls, header: str, msg: str) -> NoReturn:
        import tkinter.messagebox
        tkinter.messagebox.showerror(header, msg)
        sys.exit(msg)

    @classmethod
    def _report_fatal_error(cls, title: str, msg: str, cause: Exception) -> NoReturn:
        logger.exception(cause)
        cls.popup_error(title, msg)

    @staticmethod
    def _get_thread(name: str):
        from java.lang import Thread
        for t in Thread.getAllStackTraces().keySet():
            if t.getName() == name:
                return t
        return None

    def _launch(self):
        import ctypes
        from ghidra import GhidraRun
        from java.lang import Runtime, Thread

        if sys.platform == "win32":
            appid = ctypes.c_wchar_p(self.app_info.name)
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)

        stdout = _PyhidraStdOut(sys.stdout)
        stderr = _PyhidraStdOut(sys.stderr)
        with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
            jpype.setupGuiEnvironment(lambda: GhidraRun().launch(self._layout, self.args))
            is_exiting = threading.Event()
            Runtime.getRuntime().addShutdownHook(Thread(is_exiting.set))
            try:
                is_exiting.wait()
            finally:
                jpype.shutdownGuiEnvironment()



import struct
import sys
import sysconfig
from pathlib import Path
from pyhidra import DeferredPyhidraLauncher


# creating a shortcut with the winapi to have a set app id is trivial right?


def create_shortcut(link: Path):
    if not link.is_absolute():
        link = link.absolute()
    link = link.with_suffix(".lnk")
    if link.exists():
        sys.exit(f"{link} already exists")

    import ctypes
    import ctypes.wintypes

    class _GUID(ctypes.wintypes.DWORD * 4):
        def __init__(self, guid: str) -> None:
            ctypes.oledll.ole32.CLSIDFromString(guid, ctypes.byref(self))

    class _PROPERTYKEY(ctypes.wintypes.DWORD * 5):
        def __init__(self, key: str, pid: int) -> None:
            ctypes.oledll.ole32.IIDFromString(key, ctypes.byref(self))
            self[-1] = pid

    launcher = DeferredPyhidraLauncher()

    _PropertyVariant = struct.Struct(f"B7xP{ctypes.sizeof(ctypes.c_void_p())}x")
    _AppUserModelId = _PROPERTYKEY("{9F4C2855-9F79-4B39-A8D0-E1D42DE1D5F3}", 5)
    _CLSID_ShellLink = _GUID("{00021401-0000-0000-C000-000000000046}")
    _IID_IShellLinkW = _GUID("{000214F9-0000-0000-C000-000000000046}")
    _IID_IPersistFile = _GUID("{0000010B-0000-0000-C000-000000000046}")
    _IID_IPropertyStore = _GUID("{886d8eeb-8cf2-4446-8d02-cdba1dbdcf99}")

    _CLSCTX_INPROC_SERVER = 1
    _COINIT_APARTMENTTHREADED = 2
    _COINIT_DISABLE_OLE1DDE = 4
    _VT_LPWSTR = 31
    _APP_ID = launcher.app_info.name

    WINFUNCTYPE = ctypes.WINFUNCTYPE
    _CoCreateInstance = ctypes.oledll.ole32.CoCreateInstance
    _QueryInterface = WINFUNCTYPE(ctypes.HRESULT, _GUID, ctypes.c_void_p)(0, "QueryInterface")
    _Release = WINFUNCTYPE(ctypes.HRESULT)(2, "Release")
    _Save = WINFUNCTYPE(ctypes.HRESULT, ctypes.c_wchar_p, ctypes.wintypes.BOOL)(6, "Save")
    _SetPath = WINFUNCTYPE(ctypes.HRESULT, ctypes.c_wchar_p)(20, "SetPath")
    _SetDescription = WINFUNCTYPE(ctypes.HRESULT, ctypes.c_wchar_p)(7, "SetDescription")
    _SetIconLocation = WINFUNCTYPE(ctypes.HRESULT, ctypes.c_wchar_p)(17, "SetIconLocation")
    _SetValue = WINFUNCTYPE(ctypes.HRESULT, ctypes.c_void_p, ctypes.c_void_p)(6, "SetValue")

    link = str(link)
    target = Path(sysconfig.get_path("scripts")) / "pyhidraw.exe"
    icon = str(launcher.install_dir / "support" / "ghidra.ico")
    p_link = ctypes.c_void_p()
    p_file = ctypes.c_void_p()
    p_store = ctypes.c_void_p()
    p_app_id = ctypes.wintypes.LPCWSTR(_APP_ID)
    ctypes.oledll.ole32.CoInitializeEx(None, _COINIT_APARTMENTTHREADED | _COINIT_DISABLE_OLE1DDE)
    try:
        ref = ctypes.byref(p_link)
        _CoCreateInstance(_CLSID_ShellLink, None, _CLSCTX_INPROC_SERVER, _IID_IShellLinkW, ref)
        _SetPath(p_link, ctypes.c_wchar_p(str(target)))
        _SetDescription(p_link, p_app_id)
        _SetIconLocation(p_link, ctypes.c_wchar_p(icon))
        _QueryInterface(p_link, _IID_IPropertyStore, ctypes.byref(p_store))
        value = _PropertyVariant.pack(_VT_LPWSTR, ctypes.cast(p_app_id, ctypes.c_void_p).value)
        value = (ctypes.c_byte * len(value))(*value)
        _SetValue(p_store, ctypes.byref(_AppUserModelId), ctypes.byref(value))
        _QueryInterface(p_link, _IID_IPersistFile, ctypes.byref(p_file))
        _Save(p_file, ctypes.c_wchar_p(link), True)
    finally:
        if p_file:
            _Release(p_file)
        if p_link:
            _Release(p_link)
        if p_store:
            _Release(p_store)
        ctypes.oledll.ole32.CoUninitialize()




from pathlib import Path
import textwrap
import importlib
import sys
import jpype
import pyhidra
import pytest

EXE_NAME = "strings.exe"


def test_invalid_jpype_keyword_arg():
    assert not jpype.isJVMStarted()

    launcher = pyhidra.launcher.HeadlessPyhidraLauncher()
    with pytest.raises(TypeError) as ex:
        launcher.start(someBogusKeywordArg=True)
    assert "startJVM() got an unexpected keyword argument 'someBogusKeywordArg'" in str(ex.value)


def test_invalid_vm_arg_succeed():
    assert not jpype.isJVMStarted()

    launcher = pyhidra.launcher.HeadlessPyhidraLauncher()
    launcher.add_vmargs('-XX:SomeBogusJvmArg')
    launcher.start(ignoreUnrecognized=True)


def test_run_script(capsys, shared_datadir: Path):
    strings_exe = shared_datadir / EXE_NAME
    script_path = shared_datadir / "example_script.py"
    pyhidra.run_script(strings_exe, script_path, script_args=["my", "--commands"], analyze=False)
    captured = capsys.readouterr()

    expected = textwrap.dedent(f"""\
        {script_path} my --commands
        my --commands
        {EXE_NAME} - .ProgramDB
    """)

    assert captured.out == expected


def test_open_program(shared_datadir: Path):
    strings_exe = shared_datadir / EXE_NAME
    with pyhidra.open_program(strings_exe, analyze=False, language="x86:LE:32:default", compiler="windows") as flat_api:
        assert flat_api.currentProgram.name == strings_exe.name
        assert flat_api.getCurrentProgram().listing
        assert flat_api.getCurrentProgram().changeable


def test_bad_language(shared_datadir: Path):
    strings_exe = shared_datadir / EXE_NAME
    with pytest.raises(ValueError):
        with pyhidra.open_program(
            strings_exe,
            analyze=False,
            language="invalid",
            compiler="windows"
        ) as flat_api:
            pass


def test_bad_compiler(shared_datadir: Path):
    strings_exe = shared_datadir / EXE_NAME
    with pytest.raises(ValueError):
        with pyhidra.open_program(
            strings_exe,
            analyze=False,
            language="x86:LE:32:default",
            compiler="invalid"
        ) as flat_api:
            pass


def test_no_compiler(shared_datadir: Path):
    strings_exe = shared_datadir / EXE_NAME
    with pyhidra.open_program(strings_exe, analyze=False, language="x86:LE:32:default") as flat_api:
        pass


def test_no_language_with_compiler(shared_datadir: Path):
    strings_exe = shared_datadir / EXE_NAME
    with pyhidra.open_program(strings_exe, analyze=False, compiler="windows") as flat_api:
        pass


def test_no_project(capsys, shared_datadir: Path):
    pyhidra.run_script(None, shared_datadir / "projectless_script.py")
    captured = capsys.readouterr()
    assert captured.out.rstrip() == "projectless_script executed successfully"


def test_no_program(capsys, shared_datadir: Path):
    script_path = shared_datadir / "programless_script.py"
    project_path = shared_datadir / "programless_ghidra"

    pyhidra.run_script(None, script_path, project_path, "programless")
    captured = capsys.readouterr()
    assert captured.out.rstrip() == "programless_script executed successfully"


def test_import_script(capsys, shared_datadir: Path):
    script_path = shared_datadir / "import_test_script.py"
    pyhidra.run_script(None, script_path)
    captured = capsys.readouterr()
    assert captured.out.rstrip() == "imported successfully"


def test_import_ghidra_base_java_packages():

    def get_runtime_top_level_java_packages(launcher) -> set:
        from java.lang import Package

        packages = set()

        # Applicaiton needs to fully intialize to find all Ghidra packages
        if launcher.has_launched():

            for package in Package.getPackages():
                # capture base packages only
                packages.add(package.getName().split('.')[0])

        # Remove dc3 base package as it doesn't exist and won't conflict
        packages.remove('dc3')

        return packages

    def wrap_mod(mod):
        return mod + '_'

    launcher = pyhidra.start()

    # Test to ensure _PyhidraImportLoader is last loader
    assert isinstance(sys.meta_path[-1], pyhidra.launcher._PyhidraImportLoader)

    packages = get_runtime_top_level_java_packages(launcher)

    assert len(packages) > 0

    # Test full coverage for Java base packages (_JImportLoader or _PyhidraImportLoader)
    for mod in packages:
        # check spec using standard import machinery "import mod"
        spec = importlib.util.find_spec(mod)
        if not isinstance(spec.loader, jpype.imports._JImportLoader):
            # handle case with conflict. check spec with "import mod_"
            spec = importlib.util.find_spec(wrap_mod(mod))

        assert spec is not None
        assert isinstance(spec.loader, jpype.imports._JImportLoader) or isinstance(
            spec.loader, pyhidra.launcher._PyhidraImportLoader)

    # Test all Java base packages are available with '_'
    for mod in packages:
        spec_ = importlib.util.find_spec(wrap_mod(mod))
        assert spec_ is not None
        assert isinstance(spec_.loader, pyhidra.launcher._PyhidraImportLoader)

    # Test standard import
    import ghidra
    assert isinstance(ghidra.__loader__, jpype.imports._JImportLoader)

    # Test import with conflict
    import pdb_
    assert isinstance(pdb_.__loader__, pyhidra.launcher._PyhidraImportLoader)

    # Test "from" import with conflict
    from pdb_ import PdbPlugin
    from pdb_.symbolserver import LocalSymbolStore

    # Test _Jpackage handles import that doesn't exist
    try:
        import pdb_.doesntexist
    except ImportError:
        pass



import keyword
import logging

import jpype


# pylint: disable=no-member, too-few-public-methods
@jpype.JImplementationFor("java.lang.Object")
class _JavaObject:

    def __jclass_init__(self: jpype.JClass):
        try:
            if isinstance(self, jpype.JException):
                # don't process any exceptions
                return
            exposer = jpype.JClass("dc3.pyhidra.plugin.PythonFieldExposer")
            if exposer.class_.isAssignableFrom(self.class_):
                return
            utils = jpype.JClass("dc3.pyhidra.property.PropertyUtils")
            for prop in utils.getProperties(self.class_):
                field = prop.field
                if keyword.iskeyword(field):
                    field += '_'
                if field == "class_":
                    continue
                fget = prop.fget if prop.hasGetter() else None
                fset = prop.fset if prop.hasSetter() else None
                self._customize(field, property(fget, fset))

        # allowing any exception to escape here may cause the jvm to terminate
        # pylint: disable=bare-except
        except:
            logger = logging.getLogger(__name__)
            logger.error("Failed to add property customizations for %s", self, exc_info=1)

    def __repr__(self):
        return str(self)




__version__ = "1.0.0"

# Expose API
from .core import run_script, start, started, open_program
from .launcher import DeferredPyhidraLauncher, GuiPyhidraLauncher, HeadlessPyhidraLauncher
from .script import get_current_interpreter
from .version import ApplicationInfo, ExtensionDetails


__all__ = [
    "get_current_interpreter", "open_program", "run_script", "start", "started",
    "ApplicationInfo", "DeferredPyhidraLauncher", "ExtensionDetails",
    "GuiPyhidraLauncher", "HeadlessPyhidraLauncher"
]




import logging
import os
import platform
import sys
import warnings

from pyhidra import get_current_interpreter as _get_current_interpreter

logger = logging.getLogger(__name__)


def _gui():
    if platform.system() == 'Windows':
        # gui_script works like it is supposed to on windows
        gui()
        return

    pid = os.fork()
    if pid != 0:
        # original process can exit
        return

    fd = os.open(os.devnull, os.O_RDWR)
    # redirect stdin, stdout and stderr to /dev/null so the jvm can't use the terminal
    # this also prevents errors from attempting to write to a closed sys.stdout #21
    os.dup2(fd, sys.stdin.fileno(), inheritable=False)
    os.dup2(fd, sys.stdout.fileno(), inheritable=False)
    os.dup2(fd, sys.stderr.fileno(), inheritable=False)

    # run the application
    gui()


def gui():
    """
    Starts the Ghidra GUI
    """
    if "GHIDRA_INSTALL_DIR" not in os.environ:
        import tkinter.messagebox
        msg = "GHIDRA_INSTALL_DIR not set.\nPlease see the README for setup instructions"
        tkinter.messagebox.showerror("Improper Setup", msg)
        sys.exit()
    import pyhidra
    pyhidra.GuiPyhidraLauncher().start()


def get_current_interpreter():
    warnings.warn(
        "get_current_interpreter has been moved. Please use pyhidra.get_current_interpreter",
        DeprecationWarning
    )
    return _get_current_interpreter()




import argparse
import code

import sys
from pathlib import Path

import pyhidra
import pyhidra.core
import pyhidra.gui


def _create_shortcut():
    from pyhidra.win_shortcut import create_shortcut
    create_shortcut(Path(sys.argv[-1]))


def _interpreter(interpreter_globals: dict):
    from ghidra.framework import Application
    version = Application.getApplicationVersion()
    name = Application.getApplicationReleaseName()
    banner = f"Python Interpreter for Ghidra {version} {name}\n"
    banner += f"Python {sys.version} on {sys.platform}"
    code.interact(banner=banner, local=interpreter_globals, exitmsg='')


# pylint: disable=too-few-public-methods
class PyhidraArgs(argparse.Namespace):
    """
    Custom namespace for holding the command line arguments
    """

    def __init__(self, parser: argparse.ArgumentParser, **kwargs):
        super().__init__(**kwargs)
        self.parser = parser
        self.valid = True
        self.verbose = False
        self.skip_analysis = False
        self.binary_path: Path = None
        self.script_path: Path = None
        self.project_name = None
        self.project_path: Path = None
        self.install_dir: Path = None
        self._script_args = []

    def func(self):
        """
        Run script or enter repl
        """
        if not self.valid:
            self.parser.print_usage()
            return

        if self.script_path is not None:
            try:
                pyhidra.run_script(
                    self.binary_path,
                    self.script_path,
                    project_location=self.project_path,
                    project_name=self.project_name,
                    script_args=self._script_args,
                    verbose=self.verbose,
                    analyze=not self.skip_analysis,
                    install_dir=self.install_dir
                )
            except KeyboardInterrupt:
                # gracefully finish when cancelled
                pass
        elif self.binary_path is not None:
            args = (
                self.binary_path,
                self.project_path,
                self.project_name,
                self.verbose,
                not self.skip_analysis
            )
            with pyhidra.core._flat_api(*args, install_dir=self.install_dir) as api:
                _interpreter(api)
        else:
            pyhidra.HeadlessPyhidraLauncher(
                verbose=self.verbose, install_dir=self.install_dir).start()
            _interpreter(globals())

    @property
    def script_args(self):
        return self._script_args

    @script_args.setter
    def script_args(self, value):
        if self._script_args is None:
            self._script_args = value
        else:
            # append any remaining args to the ones which were previously consumed
            self._script_args.extend(value)


class PathAction(argparse.Action):
    """
    Custom action for handling script and binary paths as positional arguments
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nargs = '*'
        self.type = str

    def __call__(self, parser, namespace: PyhidraArgs, values, option_string=None):

        if not values:
            return

        if namespace.script_path is not None:
            # Any arguments after the script path get passed to the script
            namespace.script_args = values
            return

        value = Path(values.pop(0))

        if not value.exists():
            # File must exist
            namespace.valid = False

        if value.suffix == ".py":
            namespace.script_path = value
            namespace.script_args = values
            return

        if namespace.binary_path is None:
            # Peek at the next value, if present, to check if it is a script
            # The optional binary file MUST come before the script
            if len(values) > 0 and not values[0].endswith(".py"):
                namespace.valid = False

            namespace.binary_path = value

        if not values:
            return

        # Recurse until all values are consumed
        # The remaining arguments in the ArgParser was a lie for pretty help text
        # and to pick up trailing optional arguments meant for the script
        self(parser, namespace, values)


def _get_parser():
    is_win32 = sys.platform == "win32"
    extra_flag = "[-s] " if is_win32 else ""
    usage = f"pyhidra [-h] [-v] [-g] {extra_flag}[--no-analysis] [--project-name name] [--project-path path] " \
            "[binary_path] [script_path] ..."
    parser = argparse.ArgumentParser(prog="pyhidra", usage=usage)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Enable verbose output during Ghidra initialization"
    )
    parser.add_argument(
        "-g",
        "--gui",
        action="store_const",
        dest="func",
        const=pyhidra.gui.gui,
        help="Start Ghidra GUI"
    )
    if is_win32:
        parser.add_argument(
            "-s",
            "--shortcut",
            action="store_const",
            dest="func",
            const=_create_shortcut,
            help="Creates a shortcut that can be pinned to the taskbar (Windows only)"
        )
    parser.add_argument(
        "--install-dir",
        type=Path,
        default=None,
        dest="install_dir",
        metavar="",
        help="Path to Ghidra installation. "\
             "(defaults to the GHIDRA_INSTALL_DIR environment variable)"
    )
    parser.add_argument(
        "--skip-analysis",
        dest="skip_analysis",
        action="store_true",
        help="Switch to skip analysis after loading the binary file if provided"
    )
    parser.add_argument(
        "binary_path",
        action=PathAction,
        help="Optional binary path"
    )
    parser.add_argument(
        "script_path",
        action=PathAction,
        help=(
            "Headless script path. The script must have a .py extension. "
            "If a script is not provided, pyhidra will drop into a repl."
        )
    )
    parser.add_argument(
        "script_args",
        help="Arguments to be passed to the headless script",
        nargs=argparse.REMAINDER
    )
    parser.add_argument(
        "--project-name",
        type=str,
        dest="project_name",
        metavar="name",
        help="Project name to use. "
             "(defaults to binary filename with \"_ghidra\" suffix if provided else None)"
    )
    parser.add_argument(
        "--project-path",
        type=Path,
        dest="project_path",
        metavar="path",
        help="Location to store project. "
             "(defaults to same directory as binary file if provided else None)"
    )
    return parser


def main():
    """
    pyhidra module main function
    """
    parser = _get_parser()
    parser.parse_args(namespace=PyhidraArgs(parser)).func()


if __name__ == "__main__":
    main()



import contextlib
from pathlib import Path
from typing import Union, TYPE_CHECKING, Tuple, ContextManager, List, Optional

from pyhidra.converters import *  # pylint: disable=wildcard-import, unused-wildcard-import


if TYPE_CHECKING:
    from pyhidra.launcher import PyhidraLauncher
    from ghidra.base.project import GhidraProject
    from ghidra.program.flatapi import FlatProgramAPI
    from ghidra.program.model.lang import CompilerSpec, Language, LanguageService
    from ghidra.program.model.listing import Program


def start(verbose=False) -> "PyhidraLauncher":
    """
    Starts the JVM and loads the Ghidra libraries.
    Full Ghidra initialization is deferred.

    :param verbose: Enable verbose output during JVM startup (Defaults to False)
    :return: The DeferredPhyidraLauncher used to start the JVM
    """
    from pyhidra.launcher import HeadlessPyhidraLauncher
    launcher = HeadlessPyhidraLauncher(verbose=verbose)
    launcher.start()
    return launcher


def started() -> bool:
    """
    Whether the PyhidraLauncher has already started.
    """
    from pyhidra.launcher import PyhidraLauncher
    return PyhidraLauncher.has_launched()


def _get_language(id: str) -> "Language":
    from ghidra.program.util import DefaultLanguageService
    from ghidra.program.model.lang import LanguageID, LanguageNotFoundException
    try:
        service: "LanguageService" = DefaultLanguageService.getLanguageService()
        return service.getLanguage(LanguageID(id))
    except LanguageNotFoundException:
        # suppress the java exception
        pass
    raise ValueError("Invalid Language ID: "+id)


def _get_compiler_spec(lang: "Language", id: str = None) -> "CompilerSpec":
    if id is None:
        return lang.getDefaultCompilerSpec()
    from ghidra.program.model.lang import CompilerSpecID, CompilerSpecNotFoundException
    try:
        return lang.getCompilerSpecByID(CompilerSpecID(id))
    except CompilerSpecNotFoundException:
        # suppress the java exception
        pass
    lang_id = lang.getLanguageID()
    raise ValueError(f"Invalid CompilerSpecID: {id} for Language: {lang_id.toString()}")


def _setup_project(
        binary_path: Union[str, Path],
        project_location: Union[str, Path] = None,
        project_name: str = None,
        language: str = None,
        compiler: str = None
) -> Tuple["GhidraProject", "Program"]:
    from ghidra.base.project import GhidraProject
    from java.io import IOException
    if binary_path is not None:
        binary_path = Path(binary_path)
    if project_location:
        project_location = Path(project_location)
    else:
        project_location = binary_path.parent
    if not project_name:
        project_name = f"{binary_path.name}_ghidra"
    project_location /= project_name
    project_location.mkdir(exist_ok=True, parents=True)

    # Open/Create project
    program: "Program" = None
    try:
        project = GhidraProject.openProject(project_location, project_name, True)
        if binary_path is not None:
            if project.getRootFolder().getFile(binary_path.name):
                program = project.openProgram("/", binary_path.name, False)
    except IOException:
        project = GhidraProject.createProject(project_location, project_name, False)

    if binary_path is not None and program is None:
        if language is None:
            program = project.importProgram(binary_path)
            if program is None:
                raise RuntimeError(f"Ghidra failed to import '{binary_path}'. Try providing a language manually.")
        else:
            lang = _get_language(language)
            comp = _get_compiler_spec(lang, compiler)
            program = project.importProgram(binary_path, lang, comp)
            if program is None:
                message = f"Ghidra failed to import '{binary_path}'. "
                if compiler:
                    message += f"The provided language/compiler pair ({language} / {compiler}) may be invalid."
                else:
                    message += f"The provided language ({language}) may be invalid."
                raise ValueError(message)
        project.saveAs(program, "/", program.getName(), True)

    return project, program


def _setup_script(project: "GhidraProject", program: "Program"):
    from pyhidra.script import PyGhidraScript
    from ghidra.app.script import GhidraState
    from ghidra.program.util import ProgramLocation
    from ghidra.util.task import TaskMonitor

    from java.io import PrintWriter
    from java.lang import System

    if project is not None:
        project = project.getProject()

    location = None
    if program is not None:
        # create a GhidraState and setup a HeadlessScript with it
        mem = program.getMemory().getLoadedAndInitializedAddressSet()
        if not mem.isEmpty():
            location = ProgramLocation(program, mem.getMinAddress())
    state = GhidraState(None, project, program, location, None, None)
    script = PyGhidraScript()
    script.set(state, TaskMonitor.DUMMY, PrintWriter(System.out))
    return script


def _analyze_program(flat_api, program):
    from ghidra.program.util import GhidraProgramUtilities
    from ghidra.app.script import GhidraScriptUtil
    if GhidraProgramUtilities.shouldAskToAnalyze(program):
        GhidraScriptUtil.acquireBundleHostReference()
        try:
            flat_api.analyzeAll(program)
            if hasattr(GhidraProgramUtilities, "markProgramAnalyzed"):
                GhidraProgramUtilities.markProgramAnalyzed(program)
            else:
                GhidraProgramUtilities.setAnalyzedFlag(program, True)
        finally:
            GhidraScriptUtil.releaseBundleHostReference()


@contextlib.contextmanager
def open_program(
        binary_path: Union[str, Path],
        project_location: Union[str, Path] = None,
        project_name: str = None,
        analyze=True,
        language: str = None,
        compiler: str = None,
) -> ContextManager["FlatProgramAPI"]:
    """
    Opens given binary path in Ghidra and returns FlatProgramAPI object.

    :param binary_path: Path to binary file, may be None.
    :param project_location: Location of Ghidra project to open/create.
        (Defaults to same directory as binary file)
    :param project_name: Name of Ghidra project to open/create.
        (Defaults to name of binary file suffixed with "_ghidra")
    :param analyze: Whether to run analysis before returning.
    :param language: The LanguageID to use for the program.
        (Defaults to Ghidra's detected LanguageID)
    :param compiler: The CompilerSpecID to use for the program. Requires a provided language.
        (Defaults to the Language's default compiler)
    :return: A Ghidra FlatProgramAPI object.
    :raises ValueError: If the provided language or compiler is invalid.
    """

    from pyhidra.launcher import PyhidraLauncher, HeadlessPyhidraLauncher

    if not PyhidraLauncher.has_launched():
        HeadlessPyhidraLauncher().start()

    from ghidra.app.script import GhidraScriptUtil
    from ghidra.program.flatapi import FlatProgramAPI

    project, program = _setup_project(
        binary_path,
        project_location,
        project_name,
        language,
        compiler
    )
    GhidraScriptUtil.acquireBundleHostReference()

    try:
        flat_api = FlatProgramAPI(program)

        if analyze:
            _analyze_program(flat_api, program)

        yield flat_api
    finally:
        GhidraScriptUtil.releaseBundleHostReference()
        project.save(program)
        project.close()


@contextlib.contextmanager
def _flat_api(
        binary_path: Union[str, Path] = None,
        project_location: Union[str, Path] = None,
        project_name: str = None,
        verbose=False,
        analyze=True,
        language: str = None,
        compiler: str = None,
        *,
        install_dir: Path = None
):
    """
    Runs a given script on a given binary path.

    :param binary_path: Path to binary file, may be None.
    :param script_path: Path to script to run.
    :param project_location: Location of Ghidra project to open/create.
        (Defaults to same directory as binary file)
    :param project_name: Name of Ghidra project to open/create.
        (Defaults to name of binary file suffixed with "_ghidra")
    :param script_args: Command line arguments to pass to script.
    :param verbose: Enable verbose output during Ghidra initialization.
    :param analyze: Whether to run analysis, if a binary_path is provided, before returning.
    :param language: The LanguageID to use for the program.
        (Defaults to Ghidra's detected LanguageID)
    :param compiler: The CompilerSpecID to use for the program. Requires a provided language.
        (Defaults to the Language's default compiler)
    :param install_dir: The path to the Ghidra installation directory. This parameter is only
        used if Ghidra has not been started yet.
        (Defaults to the GHIDRA_INSTALL_DIR environment variable)
    :raises ValueError: If the provided language or compiler is invalid.
    """
    from pyhidra.launcher import PyhidraLauncher, HeadlessPyhidraLauncher

    if not PyhidraLauncher.has_launched():
        HeadlessPyhidraLauncher(verbose=verbose, install_dir=install_dir).start()

    project, program = None, None
    if binary_path or project_location:
        project, program = _setup_project(
            binary_path,
            project_location,
            project_name,
            language,
            compiler
        )

    from ghidra.app.script import GhidraScriptUtil

    # always aquire a bundle reference to avoid a NPE when attempting to run any Java scripts
    GhidraScriptUtil.acquireBundleHostReference()
    try:
        script = _setup_script(project, program)
        if analyze and program is not None:
            _analyze_program(script, program)
        yield script
    finally:
        GhidraScriptUtil.releaseBundleHostReference()
        if project is not None:
            if program is not None:
                project.save(program)
            project.close()


# pylint: disable=too-many-arguments
def run_script(
    binary_path: Optional[Union[str, Path]],
    script_path: Union[str, Path],
    project_location: Union[str, Path] = None,
    project_name: str = None,
    script_args: List[str] = None,
    verbose=False,
    analyze=True,
    lang: str = None,
    compiler: str = None,
    *,
    install_dir: Path = None
):
    """
    Runs a given script on a given binary path.

    :param binary_path: Path to binary file, may be None.
    :param script_path: Path to script to run.
    :param project_location: Location of Ghidra project to open/create.
        (Defaults to same directory as binary file if None)
    :param project_name: Name of Ghidra project to open/create.
        (Defaults to name of binary file suffixed with "_ghidra" if None)
    :param script_args: Command line arguments to pass to script.
    :param verbose: Enable verbose output during Ghidra initialization.
    :param analyze: Whether to run analysis, if a binary_path is provided, before running the script.
    :param lang: The LanguageID to use for the program.
        (Defaults to Ghidra's detected LanguageID)
    :param compiler: The CompilerSpecID to use for the program. Requires a provided language.
        (Defaults to the Language's default compiler)
    :param install_dir: The path to the Ghidra installation directory. This parameter is only
        used if Ghidra has not been started yet.
        (Defaults to the GHIDRA_INSTALL_DIR environment variable)
    :raises ValueError: If the provided language or compiler is invalid.
    """
    script_path = str(script_path)
    args = binary_path, project_location, project_name, verbose, analyze, lang, compiler
    with _flat_api(*args, install_dir=install_dir) as script:
        script.run(script_path, script_args)



"""
Script to uninstall a Ghidra plugin.
"""

import argparse
from pathlib import Path

import pyhidra


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Uninstall Ghidra Plugin")
    parser.add_argument(
        "PLUGIN_NAME",
        help="Name of plugin to uninstall"
    )
    parser.add_argument(
        "--install-dir",
        type=Path,
        default=None,
        dest="install_dir",
        metavar="",
        help="Path to Ghidra installation. "\
             "(defaults to the GHIDRA_INSTALL_DIR environment variable)"
    )
    args = parser.parse_args()

    plugin_name = args.PLUGIN_NAME
    launcher = pyhidra.HeadlessPyhidraLauncher(install_dir=args.install_dir)
    install_path = launcher.get_install_path(plugin_name)
    if install_path.exists():
        launcher.uninstall_plugin(plugin_name)
        if install_path.exists():
            print(f"Failed to uninstall {plugin_name}")
        else:
            print(f"{plugin_name} has been uninstalled.")
    else:
        print(f"{plugin_name} not installed.")



# Examples of Pyhidra-specific functionality
# @category: Examples.Python


# we can import java libraries just as if they were python libraries
from java.util import LinkedList

# and then use them like they are natural classes
java_list = LinkedList([1,2,3])
print(f"linked list object class: {java_list.__class__}")

# importing and using Ghidra modules is the same
from ghidra.program.flatapi import FlatProgramAPI
print(f"max references to a flat program api: {FlatProgramAPI.MAX_REFERENCES_TO}")

# we can also do normal python-ish things on our Java objects, like:
# indexing
print(f"first element of the list: {java_list[0]}")

# slicing
print(f"first two elements of the list: {java_list[0:2]}")

# list comprehension
java_list_double = [i * 2 for i in java_list]
print(f"list comprehension result: {java_list_double}")

# automatic calls to getters
print(f"current program name: {currentProgram.name}") # calls currentProgram.getName()

# here's an example of how this stuff might come in handy with Ghidra:
print('current program memory blocks:\n')
for block in currentProgram.memory.blocks:
    print(block.name)


# many Ghidra functions need a Java-native array to pass or receive values
# JPype provides objects of JByte, JChar, etc. to meet this need
# this example demonstrates how you would create an array of bytes to get
# the first 10 bytes of memory from the .text section

# we need this import to get at the helper classes
import jpype

# get the block we need
block = currentProgram.memory.getBlock('.text')
if block:
    # the verbose way of getting the array
    byte_array_maker = jpype.JArray(jpype.JByte)
    byte_array = byte_array_maker(10)

    # we also could have taken a shortcut with just:
    # byte_array = jpype.JByte[10]

    # let's have a look at our new object
    print(f"array class: {byte_array.__class__}")
    # will be <java class 'byte[]'>
    print(f"array length: {len(byte_array)}")

    # we can now use this array wherever a Java method requires a byte[] type
    # the signature of getBytes is getBytes(Address addr, byte[] b)
    block.getBytes(block.start, byte_array)

    # after the call, we can get the bytes out as desired
    # we just put them in a list comprehension here
    print(f"first 10 bytes of .text: {['%#x' % ((b+256)%256) for b in byte_array]}")

else:
    print('no block named .text in this program.')

# see the user manual of JPype for more details on interoperability:
# https://jpype.readthedocs.io/en/latest/userguide.html
